/*
 *  worldclock home widget for the maemo desktop.
 *  Copyright (C) 2010 Nicolai Hess
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <gtk/gtk.h>

#include <math.h>
#include <libintl.h>
#include <locale.h>
#include <string.h>
#include <clockd/libtime.h>
#include <hildon/hildon.h>
#include <libhildondesktop/libhildondesktop.h>
#include "worldclock-home-widget.h"	

HD_DEFINE_PLUGIN_MODULE(WorldClockHomePlugin, world_clock_home_plugin, HD_TYPE_HOME_PLUGIN_ITEM)


#define WORLD_CLOCK_HOME_PLUGIN_SETTINGS_FILE "/.worldclock_home_widget"

static void
world_clock_get_time(int* hour, int* min, int* sec, Cityinfo* city_info)
{
  struct tm time;
  time_get_remote(time_get_time(),
		  cityinfo_get_zone(city_info),
		  &time);
  *sec = time.tm_sec;
  *min = time.tm_min;
  *hour = time.tm_hour;
}

void
world_clock_save_settings(WorldClockHomePlugin* desktop_plugin)
{
  GKeyFile *keyFile;
  gchar *fileData;
  FILE *iniFile;
  gsize size;
  gchar *filename;

  keyFile = g_key_file_new();
  filename = g_strconcat (g_get_home_dir(), WORLD_CLOCK_HOME_PLUGIN_SETTINGS_FILE, NULL);
  g_key_file_load_from_file (keyFile, filename, G_KEY_FILE_KEEP_COMMENTS, NULL);
  g_key_file_set_boolean (keyFile, desktop_plugin->iD, "collapsed", desktop_plugin->collapsed);
  g_key_file_set_integer (keyFile, desktop_plugin->iD, "zone_id", desktop_plugin->city_info_id);
  
  fileData = g_key_file_to_data (keyFile, &size, NULL);
  g_file_set_contents(filename, fileData, size, NULL);

  g_key_file_free (keyFile); 
  g_free(fileData); 
  g_free(filename); 
}

gchar* 
world_clock_gmt_offset_string(Cityinfo* cityinfo)
{
  int offset = time_get_utc_offset(cityinfo_get_zone(cityinfo));
  if((offset % 3600)==0)
    return g_strdup_printf("UTC %+d", -offset/3600);
  else
    return g_strdup_printf("UTC %+d:%d", -offset/3600, 30);
}

void 
init_clock_list(GtkListStore* clock_list)
{
  Cityinfo** cities = cityinfo_get_all();
  Cityinfo** cities_iter = cities;
  GtkTreeIter iter;
  while((*cities_iter))
  {
    gchar* name = cityinfo_get_name(*cities_iter);
    gchar* country = cityinfo_get_country(*cities_iter);
    gchar* text = g_strdup_printf("%s, %s (%s)", name, country, world_clock_gmt_offset_string(*cities_iter));
    gint id = cityinfo_get_id(*cities_iter);
    
    if(name)
    {
      gtk_list_store_append (clock_list, &iter); 
      gtk_list_store_set(clock_list, &iter, 
			 0, text, 
			 1, id, -1);

    }
    ++cities_iter;
  }
  cityinfo_free_all(cities);
}

int clock_city_name_sort_func(GtkTreeModel *model,
			      GtkTreeIter *a,
			      GtkTreeIter *b,
			      gpointer user_data)
{
  gchar* val1;
  gchar* val2;
  gtk_tree_model_get(model,
		     a, 0, &val1, -1);
  gtk_tree_model_get(model,
		     b, 0, &val2, -1);
  int ret = g_strcmp0(val1, val2);
  g_free(val1);
  g_free(val2);
  return ret;
}

static void
world_clock_show_settings_dialog(GtkWidget* widget, gpointer data)
{
  WorldClockHomePlugin *desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(widget);

  GtkWidget *dialog;

  GtkCellRenderer *renderer = NULL;
  GtkWidget* selector;
  GtkListStore *clock_list = NULL;

  HildonTouchSelectorColumn* column = NULL;
  clock_list = gtk_list_store_new(2,
				  G_TYPE_STRING,
				  G_TYPE_INT);
  gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(clock_list),
				       0,
				       GTK_SORT_ASCENDING);

  gtk_tree_sortable_set_sort_func(GTK_TREE_SORTABLE(clock_list),
				  0,
				  clock_city_name_sort_func,
				  NULL,
				  NULL);
  init_clock_list(clock_list);
  GtkWidget* clockButton = hildon_picker_button_new(HILDON_SIZE_AUTO, 
						    HILDON_BUTTON_ARRANGEMENT_VERTICAL);
  hildon_button_set_title(HILDON_BUTTON(clockButton),
			  dgettext("osso-clock", "dati_ia_pr_timezone"));
  hildon_button_set_value(HILDON_BUTTON(clockButton),
			  g_strdup(cityinfo_get_name(desktop_plugin->city_info)));
  GtkWidget* displayNameEntry = hildon_entry_new(HILDON_SIZE_FINGER_HEIGHT | HILDON_SIZE_AUTO_WIDTH);


  dialog = gtk_dialog_new_with_buttons("Settings",
				       NULL,
				       0,
				       dgettext("hildon-libs", "wdgt_bd_done"),
				       GTK_RESPONSE_ACCEPT,
				       NULL);
  //  gtk_window_set_default_size(GTK_WINDOW(dialog), -1, 300);
  
  selector = hildon_touch_selector_entry_new_text();
  hildon_touch_selector_set_model(HILDON_TOUCH_SELECTOR(selector),
				  0, GTK_TREE_MODEL(clock_list));

  hildon_touch_selector_entry_set_text_column(HILDON_TOUCH_SELECTOR_ENTRY(selector), 0);
  hildon_picker_button_set_selector(HILDON_PICKER_BUTTON(clockButton),
				    HILDON_TOUCH_SELECTOR(selector));


  GtkWidget* clockNameBox = gtk_hbox_new(TRUE, 3);


  gtk_box_pack_start(GTK_BOX(clockNameBox), clockButton, TRUE, TRUE, 3);

  GtkWidget* content = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
  gtk_box_pack_start(GTK_BOX(content), clockNameBox, FALSE, FALSE, 3);
  gtk_widget_show_all(dialog);

  if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_ACCEPT)
  {
    GtkTreeIter iter;
    hildon_touch_selector_get_selected(HILDON_TOUCH_SELECTOR(selector),
				       0,
				       &iter);
    if(gtk_list_store_iter_is_valid(clock_list, &iter))
    {
      gint id = 0;
      gtk_tree_model_get(GTK_TREE_MODEL(clock_list),
			 &iter, 1, &id, -1);
      if(desktop_plugin->city_info)
      {
	cityinfo_free(desktop_plugin->city_info);
      }
      desktop_plugin->city_info = cityinfo_from_id(id);
      desktop_plugin->city_info_id = id;
    
      gtk_widget_queue_draw(GTK_WIDGET(desktop_plugin));
      world_clock_save_settings(desktop_plugin);
    }
  }
  gtk_widget_destroy (dialog);
  g_object_unref(clock_list);
}

void 
read_settings(WorldClockHomePlugin *desktop_plugin)
{
  GKeyFile *keyFile;
  gchar* fileName;
  gboolean fileExists;

  keyFile = g_key_file_new();
  fileName = g_strconcat(g_get_home_dir(), WORLD_CLOCK_HOME_PLUGIN_SETTINGS_FILE, NULL);
  fileExists = g_key_file_load_from_file (keyFile, fileName, G_KEY_FILE_KEEP_COMMENTS, NULL);
  if(fileExists)
  {

    GError *error = NULL;
    desktop_plugin->collapsed = g_key_file_get_boolean(keyFile, desktop_plugin->iD, "collapsed", &error);
    desktop_plugin->city_info_id = g_key_file_get_integer (keyFile, desktop_plugin->iD, "zone_id", &error);
    if(error)
    {
      g_error_free(error);
      error = NULL;
    }
    
    desktop_plugin->city_info = cityinfo_from_id(desktop_plugin->city_info_id);
  }
  g_key_file_free(keyFile);
  g_free(fileName);
}

gboolean
world_clock_button_release(GtkWidget* widget, GdkEventButton *event, WorldClockHomePlugin* desktop_plugin)
{
  if((event->x<80 && event->y <80) &&
     ((event->type & GDK_BUTTON_RELEASE) == GDK_BUTTON_RELEASE))
  {
    gboolean collapsed = desktop_plugin->collapsed;
    if(collapsed)
    {
      desktop_plugin->collapsed = FALSE;
    }
    else
    {
      desktop_plugin->collapsed = TRUE;
    }
    world_clock_save_settings(desktop_plugin);
    gtk_widget_queue_draw(GTK_WIDGET(desktop_plugin));
  }
  return FALSE;
}

gboolean
world_clock_update_time(gpointer data)
{
  WorldClockHomePlugin* desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(data);
  gtk_widget_queue_draw(GTK_WIDGET(desktop_plugin));

  int hour, min, sec;
  world_clock_get_time(&hour, &min, &sec, desktop_plugin->city_info);
  //  if(sec<50)
  //  {
    if(desktop_plugin->time_out_handler)
    {
      g_source_remove(desktop_plugin->time_out_handler);
      desktop_plugin->time_out_handler = g_timeout_add_seconds(60-sec, 
							       world_clock_update_time, 
							       (gpointer) desktop_plugin);
    }
    return FALSE;
    //  }
    //  return TRUE;
}

static void
handle_time_change(DBusGProxy *object, gint time, WorldClockHomePlugin* desktop_plugin)
{
  world_clock_update_time(desktop_plugin);
}

static void
world_clock_on_current_desktop(GtkWidget* widget,
			       gpointer data)
{
  WorldClockHomePlugin* desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(widget);
  gboolean on;
  g_object_get(widget, "is-on-current-desktop", &on, NULL);
  if(on)
  {
    world_clock_update_time(desktop_plugin);
    if(!desktop_plugin->time_out_handler)
    {
      int hour, min, sec;
      world_clock_get_time(&hour, &min, &sec, desktop_plugin->city_info);
      desktop_plugin->time_out_handler = g_timeout_add_seconds(60-sec, 
							       world_clock_update_time, 
							       (gpointer) desktop_plugin);
    }
  }
  else if(desktop_plugin->time_out_handler)
  {
    desktop_plugin->time_out_handler = 0;
    g_source_remove(desktop_plugin->time_out_handler);
  }
}

static GtkWidget*
build_ui()
{
  GtkWidget *contents = gtk_event_box_new();
  gtk_event_box_set_visible_window(GTK_EVENT_BOX(contents), FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(contents), 0);
  gtk_widget_show_all(GTK_WIDGET(contents));
  return GTK_WIDGET(contents);
}

static void
register_dbus_signal_on_time_changed(WorldClockHomePlugin *desktop_plugin)
{
  desktop_plugin->dbus_conn = hd_home_plugin_item_get_dbus_g_connection(&desktop_plugin->hitem, 
									DBUS_BUS_SYSTEM, 
									NULL );

  if(desktop_plugin->dbus_conn)
  {
    desktop_plugin->dbus_proxy = dbus_g_proxy_new_for_name(desktop_plugin->dbus_conn,
							   CLOCKD_SERVICE,
							   CLOCKD_PATH,
							   CLOCKD_INTERFACE);
    dbus_g_proxy_add_signal(desktop_plugin->dbus_proxy,
			    CLOCKD_TIME_CHANGED,
			    G_TYPE_INT, G_TYPE_INVALID);

    dbus_g_proxy_connect_signal(desktop_plugin->dbus_proxy,
				CLOCKD_TIME_CHANGED,
				G_CALLBACK(handle_time_change), desktop_plugin, NULL);
  }
}

static void
world_clock_home_plugin_init(WorldClockHomePlugin *desktop_plugin)
{
  GtkWidget *contents = build_ui();
  desktop_plugin->city_info = NULL;
  desktop_plugin->city_info_id = CITYINFO_WRONG_ID;
  desktop_plugin->collapsed = FALSE;
  desktop_plugin->dbus_conn = NULL;
  desktop_plugin->dbus_proxy = NULL;
  gtk_widget_set_size_request(contents, 250, 85); 
  gtk_container_add(GTK_CONTAINER(desktop_plugin), contents);
  hd_home_plugin_item_set_settings (HD_HOME_PLUGIN_ITEM (desktop_plugin), TRUE);
  g_signal_connect(GTK_CONTAINER(contents), "button-release-event", G_CALLBACK(world_clock_button_release), desktop_plugin);
  g_signal_connect(desktop_plugin, "show-settings", G_CALLBACK(world_clock_show_settings_dialog), NULL);
  g_signal_connect(desktop_plugin, "notify::is-on-current-desktop", G_CALLBACK(world_clock_on_current_desktop), NULL);
  desktop_plugin->time_out_handler = g_timeout_add_seconds(60, 
							   world_clock_update_time, 
							   (gpointer) desktop_plugin);
  register_dbus_signal_on_time_changed(desktop_plugin);
}

static void
world_clock_init_with_default_timezone(WorldClockHomePlugin* desktop_plugin)
{
  int ret = 0;
  gchar current_tz[32];
  ret = time_get_timezone(&current_tz[0],
			  32);
  if(ret>0 && ret < 32)
  {
    Cityinfo** cities = cityinfo_get_all();
    Cityinfo** cities_iter = cities;
    gboolean current_found = FALSE;
    while(*cities_iter && !current_found)
    {
      gchar* tz = cityinfo_get_zone(*cities_iter);

      if(g_strcmp0(tz, current_tz+1)==0)
      {
	current_found = TRUE;
	desktop_plugin->city_info = cityinfo_clone(*cities_iter);
      }
      ++cities_iter;
    }
    cityinfo_free_all(cities);
  }
}

static void
world_clock_realize(GtkWidget* widget)
{
  GdkScreen *screen = gtk_widget_get_screen(widget);
  gtk_widget_set_colormap(widget, gdk_screen_get_rgba_colormap(screen));
  gtk_widget_set_app_paintable(widget, TRUE);
  WorldClockHomePlugin* desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(widget);
  desktop_plugin->iD = hd_home_plugin_item_get_applet_id (HD_HOME_PLUGIN_ITEM (widget));
  read_settings(desktop_plugin);
  if(desktop_plugin->city_info == NULL || desktop_plugin->city_info_id == CITYINFO_WRONG_ID)
  {
    world_clock_init_with_default_timezone(desktop_plugin);
  }
  GTK_WIDGET_CLASS(world_clock_home_plugin_parent_class)->realize(widget);
}

static void
world_clock_draw_clock_tick(cairo_t* cr,
			    gboolean night)
{
  int i;
  cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
  cairo_save(cr);
  if(night)
    cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 0.8);
  else
    cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 0.8);

  cairo_set_line_cap(cr, CAIRO_LINE_CAP_ROUND);
  cairo_set_line_width(cr, 2);
  for(i=0;i<4;++i)
  {
    cairo_move_to(cr, 0, 20);
    cairo_line_to(cr, 0, 30);
    cairo_stroke(cr);
    cairo_rotate(cr, 2*M_PI/12.0);
    cairo_move_to(cr, 0, 25);
    cairo_line_to(cr, 0, 30);
    cairo_stroke(cr);
    cairo_rotate(cr, 2*M_PI/12.0);
    cairo_move_to(cr, 0, 25);
    cairo_line_to(cr, 0, 30);
    cairo_stroke(cr);
    cairo_rotate(cr, 2*M_PI/12.0);
  }
  cairo_restore(cr);
}

static void
world_clock_draw_clock_time(cairo_t* cr, 
			    gboolean night,
			    int hour,
			    int min,
			    int sec, 
			    WorldClockHomePlugin* desktop_plugin)
{
  cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
  if(night)
    cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 0.8);
  else
    cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 0.8);

  cairo_set_line_width(cr, 3);
  cairo_set_line_cap(cr, CAIRO_LINE_CAP_ROUND);

  cairo_save(cr);
  cairo_move_to(cr, 0, 0);
  cairo_rotate(cr, 2*M_PI/12.0*hour + 2*M_PI/(12.0*60)*min);
  cairo_line_to(cr, 0, -16);
  cairo_stroke(cr);
  cairo_restore(cr);
  cairo_save(cr);
  cairo_move_to(cr, 0, 0);
  cairo_rotate(cr, 2*M_PI/60.0*min);
  cairo_line_to(cr, 0, -24);
  cairo_stroke(cr);
  cairo_restore(cr);
}

static gchar*
world_clock_get_zone_info(WorldClockHomePlugin* desktop_plugin)
{
  struct tm remote_tz;
  time_get_remote(time_get_time(),
		  cityinfo_get_zone(desktop_plugin->city_info),
		  &remote_tz);
  gchar* tz = g_new0(gchar, 32);
  time_format_time(&remote_tz,
		   "%Z",
		   tz,
		   32);
  gchar* offset = world_clock_gmt_offset_string(desktop_plugin->city_info);
  gchar* zone_info = g_strdup_printf("%s %s", 
				     tz,
				     offset);
  g_free(tz);
  g_free(offset);
  return zone_info;
}

static gchar*
world_clock_get_date(WorldClockHomePlugin* desktop_plugin)
{
  struct tm remote_tz;
  time_get_remote(time_get_time(),
		  cityinfo_get_zone(desktop_plugin->city_info),
		  &remote_tz);
  gchar* time = g_new0(gchar, 255);
  time_format_time(&remote_tz,
		   "%a %d %b %Y",
		   time,
		   255);
  return time;
}

static void
world_clock_draw_widget_background(cairo_t* cr, gboolean collapsed)
{
  int width = 100;
  if(!collapsed)
    width = 250;
  cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
  cairo_set_source_rgba(cr, 0.6, 0.6, 0.6, 1.0);
  cairo_rectangle(cr, 0, 0, width, 100);
  cairo_fill(cr);
  cairo_set_operator(cr, CAIRO_OPERATOR_OVER);

  cairo_pattern_t* linearpattern = 
    cairo_pattern_create_linear(0, 0,
			        0, 100);
  cairo_pattern_add_color_stop_rgba(linearpattern, 
				    0.0, 
				    0.8, 0.8, 0.8, 
				    0.0);
  cairo_pattern_add_color_stop_rgba(linearpattern, 
				    90.0, 
				    0.0, 0.0, 0.0, 
				    0.42);
  cairo_set_source(cr, linearpattern);
  cairo_rectangle(cr, 0, 0, width, 100);
  cairo_fill(cr);
  cairo_pattern_destroy(linearpattern);
}

static void
world_clock_print_text(cairo_t* cr, 
		       PangoLayout* pangoLayout,
		       gchar* text,
		       gint x,
		       gint y,
		       gboolean centered)
{
  pango_layout_set_text(pangoLayout, text, -1);
  int width;
  int height;
  pango_layout_get_pixel_size(pangoLayout, &width, &height);
  cairo_set_line_width(cr, height);
  if(centered)
  {
    cairo_move_to(cr, x-width/2.0, y+(height/2.0));
    cairo_line_to(cr, x+width/2.0, y+(height/2.0));
  }
  else
  {
    cairo_move_to(cr, x, y+(height/2.0));
    cairo_line_to(cr, x+width, y+(height/2.0));
  }
  cairo_set_source_rgba(cr, 0.3, 0.3, 0.3, 0.8);
  cairo_stroke(cr);
  if(centered)
    cairo_move_to(cr, x-width/2.0, y);
  else
    cairo_move_to(cr, x, y);
  cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 1.0);
  pango_cairo_show_layout(cr, pangoLayout);
}

static void
world_clock_draw_clock_background(cairo_t* cr, gboolean night)
{
  if(night)
    cairo_set_source_rgba(cr, 0.0, 0.0, 0.0, 0.8);
  else
    cairo_set_source_rgba(cr, 1.0, 1.0, 1.0, 0.8);
  cairo_arc(cr,
	    0, 0, 
	    30,
	    0, 2*M_PI);
  cairo_fill(cr);
}

static gboolean
world_clock_expose(GtkWidget* widget, GdkEventExpose *event)
{
  cairo_t *cr;
  PangoLayout *pangoLayout;
  PangoFontDescription *fontDescription;
  WorldClockHomePlugin* desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(widget);

  cr = gdk_cairo_create(GDK_DRAWABLE(widget->window));
  gdk_cairo_region(cr, event->region);
  cairo_clip(cr);

  int min = 0;
  int hour = 0;
  int sec = 0;

  world_clock_get_time(&hour, &min, &sec, desktop_plugin->city_info);

  gboolean night = (hour > 18 || hour < 6);

  cairo_translate(cr, 50, 32);
  world_clock_draw_clock_background(cr, night);
  world_clock_draw_clock_tick(cr, night);
  world_clock_draw_clock_time(cr, night, hour, min, sec, desktop_plugin);
  cairo_translate(cr, -50, -32);

  pangoLayout = pango_cairo_create_layout(cr);
  const char* font = "Sans 10";
  fontDescription = pango_font_description_from_string(font);
  pango_layout_set_font_description(pangoLayout, fontDescription);
  pango_font_description_free(fontDescription);

  pango_layout_set_width(pangoLayout, 85*1024);
  pango_layout_set_ellipsize(pangoLayout, PANGO_ELLIPSIZE_END);

  world_clock_print_text(cr,
	                 pangoLayout,
			 cityinfo_get_name(WORLD_CLOCK_HOME_PLUGIN(widget)->city_info),
			 50, 65, TRUE);
  
  if(!WORLD_CLOCK_HOME_PLUGIN(widget)->collapsed)
  {
    pango_layout_set_width(pangoLayout, -1);
    gchar* date = world_clock_get_date(desktop_plugin);
    world_clock_print_text(cr, pangoLayout, date, 90, 5, FALSE);
    g_free(date);
    world_clock_print_text(cr, pangoLayout, cityinfo_get_country(WORLD_CLOCK_HOME_PLUGIN(widget)->city_info), 90, 25, FALSE);
    gchar* zone = world_clock_get_zone_info(desktop_plugin);
    world_clock_print_text(cr, pangoLayout, zone, 90, 45, FALSE);
    g_free(zone);
  }
  cairo_destroy(cr);
  return GTK_WIDGET_CLASS(world_clock_home_plugin_parent_class)->expose_event(widget, event);
}


static void
world_clock_plugin_finalize(GObject *object)
{
  WorldClockHomePlugin *desktop_plugin = WORLD_CLOCK_HOME_PLUGIN(object);
  if(desktop_plugin->time_out_handler)
  {
    g_source_remove(desktop_plugin->time_out_handler);
    desktop_plugin->time_out_handler = 0;
  }
  if(desktop_plugin->dbus_proxy)
  {
    dbus_g_proxy_disconnect_signal(desktop_plugin->dbus_proxy,
                                   CLOCKD_TIME_CHANGED,
                                   G_CALLBACK(handle_time_change), desktop_plugin);
  }
                                   
  g_free(desktop_plugin->iD);
  cityinfo_free(desktop_plugin->city_info);
  G_OBJECT_CLASS (world_clock_home_plugin_parent_class)->finalize (object);
}

static void
world_clock_home_plugin_class_init(WorldClockHomePluginClass *klass) 
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
  widget_class->realize = world_clock_realize;
  widget_class->expose_event = world_clock_expose;

  G_OBJECT_CLASS(klass)->finalize = world_clock_plugin_finalize;
}

static void
world_clock_home_plugin_class_finalize(WorldClockHomePluginClass *class) 
{
}
